'''
******************************************* Projet Rover le robot suiveur de lignes *****************************************************
Matériel :
Raspberry pi pico microcontroleur de gestion du robot Rover
2 batterie LIPO baton 3.7 V en série 
Bluetooth avec module HC-05 externe connecté au port COM0 série
2 moteurs et réducteurs pour les roues commandés en PWM par un driver double
1 encodeurs avec signaux A & B sur roue gauche (2 X 12 impulsions par tour, A & B déphasés de 90 °)
1 barrette de 8 leds et photodiodes comme capteur de ligne, Led renvoi 1 sur fond noir mat
1 entrée de mesure batterie
1 LED rouge état ON non pilotable
1 LED bleu d'état pour commande reçue
Liste des commandes en réceptions et réponses -->   :
 b : robot renvoie la mesure de la batterie --> BAT :xx.xV
 d :  "        "   la lecture du capteur de ligne -->  BAR :XXXXXXXX
 w : arrêt du robot et initialisation du compteur de déplacement--> ok
 lxxxx  : se déplace sur une ligne de longueur xxxx  (24 = un tour de roue soit 13.5 cm)--> ok & complete si fin de parcours atteint
 chxxxx : rotation sens horaire de longueur xxxx (51 = un tour complet sur lui même) --> ok & complete si fin de parcours atteint
 caxxxx : rotation sens anti-horaire de longueur xxxx (51 = un tour complet sur lui même)--> ok & complete si fin de parcours atteint
 vxxx   : vitesse de base du robot (par défaut à 180) --> ok
 Kxxx   : gain proportionnel du robot (par défaut à 15)--> ok
 LED BLEU sur ligne et rotation
 si commande reçue erronée --> ko
 Time Out alimentation robot off 30 minutes
 vitesse en rotation fixée à 180
 ralentissement en bout de parcours de ligne

INFO :
1 tour de roue = 13.5 cm = 24 impulsions (12 X 2 encodeur A & B)
1 tour sur lui même du robot 51 impulsions compteur

'''
from machine import PWM, Pin,Timer,UART,time_pulse_us
import time


##-----configurations des ports et entrées/sorties----##
##     variables associées                            ##  


#--> configuration UART port COM0 pour le Bluetooth 115200 bauds :
uart0 = UART(0,115200)

#--> Configuration du Time OUT pour couper l'alimentation du robot :
#######

# callback du Time Out de gestion arrêt alimentation : incrémente le compteur du Time Out:
def interruption_kill(timer):
    global cpt_kill
    cpt_kill = cpt_kill + 1
kill_Pin = Pin(18, Pin.OUT) #  commande coupure alimentation avec la Pin 18 en sortie
kill_timer = Timer(mode=Timer.PERIODIC, period=600 * 1000, callback=interruption_kill)#config du timer : 600 X 1000 ms
cpt_kill = 0 # initialisation compteur du Time Out
cpt_kill_max = 3 #  temps max avant arrêt en  X 10 minutes = 30 minutes


#######

#--> configuration entrée ADC mesure batterie :
pin_bat = machine.ADC(26) # GP26 channel 0
facteur_bat = 11.1*3.3 / (65535) # facteur de conversion ADC bat 
#lecture_bat = pin_bat.read_u16() * facteur_bat

#--> LED IR détecteur de ligne :
# En vue dessus robot : pin GP13 led IR à droite à pin GP6 Led à gauche
barled = [Pin(13,Pin.IN,Pin.PULL_DOWN),Pin(12,Pin.IN,Pin.PULL_DOWN),Pin(11,Pin.IN,Pin.PULL_DOWN),Pin(10,Pin.IN,Pin.PULL_DOWN),Pin(9,Pin.IN,Pin.PULL_DOWN),
          Pin(8,Pin.IN,Pin.PULL_DOWN),Pin(7,Pin.IN,Pin.PULL_DOWN),Pin(6,Pin.IN,Pin.PULL_DOWN)]
# Réglage sensibilité Rir du niveau de IR commandé de 0 à 100 %  par 0 à 1000
Rir = 300 # Intensité IR emise par le détecteur barre de led 30% IR 0 ...1000
# sortie PWM commande intensité de l'infrarouge des LEDS
pwmBarLed = PWM(Pin(15)) # PWM OUT réglage intensité infrarouge barrière LED
pwmBarLed.freq(40000) # 40khz fréquence de la PWM 
pwmBarLed.duty_u16(Rir*2**16//1000) # intensite 30% = 300

#--> LED bleu :
pwmLED_BLEU = PWM(Pin(28))#  pin LED BLEU d'état
pwmLED_BLEU.freq(8)
pwmLED_BLEU.duty_u16(100*2**16//1000) # Intensité 0 ...1000

#-->  paramètres régulation de suivi de ligne :
vit0 =180 # vitesse de base roue 18 % (100 % est 1000)
vitdefaut =180 # vitesse par défaut du robot 
Lcorrec = 0 # erreur décallage à gauche
Rcorrec = 0 # erreur décallage à droite
KP= 15 # correction gain proportionnel moteur
KPdefaut = 15 # KP par défaut
consigne = 4.5 # consigne tajectoire( 1 à 8) 4.5 -> centre de la barrette de LED

#--> pilotage du driver des moteurs droit et gauche :
# moteur droit : R / # moteur gauche : L
DirR = Pin(4, Pin.OUT) #  pin sens rotation moteur '0' direct / '1' reverse
DirL = Pin(5, Pin.OUT)
pwmR =PWM(Pin(2))# PWM moteur droit 
pwmR.freq(40000) # 40khz fréquence de la PWM 
pwmL =PWM(Pin(3))# PWM moteur gauche
pwmL.freq(40000) # 40khz

#--> interruptions de l'encodeur (roue gauche), des signaux A et B déphasés de 90° :
# associées aux callbacks : compteur() et compteur2()
IT_Pulse_Roue_A = Pin(14,Pin.IN,Pin.PULL_DOWN) # entrée A encodeur optique roue gaughe
IT_Pulse_Roue_B = Pin(16,Pin.IN,Pin.PULL_DOWN) # entrée B encodeur optique roue gauche

#--> initailisation variables au démarrage :
compteur_roue = 0 # compteur dimpulsion de la roue gauche 24 impulsions / tour de roue
cpt = 0 # variable comparaison compteur_roue
tourRobot = 51 # un tour robot sur lui même equivaut à 51 impulsions de la roue
tourRoue = 24 # 24 impulsions / tour de roue complet est fixé par le matériel
hor ='h' # h sens horaire de la rotaion robot a anti-horaire
select = 0 # selection fonction à effectuer une ligne '1',une rotation (tourne )'2' etpas de mouvement si '0'
stop_correct = 0 # arret régulation de correction trajectoire en fin de ligne si '1'
validCommand = 0 # si commande recue valide 

LC = ['b','w','d'] # liste des commandes à un seul caractère  
# b : robot renvoie la mesure de la batterie
# d :  "        "    la lecture du capteur de ligne
# w : arrêt du robot et initialisation du compteur de déplacement


##-----Définitions callback----##
##                             ##  

     
#--> 2 callbacks qui incrémente le compteur_roue par tour de roue  des 24 impulsions  = 12 + 12  A & B :
def compteur(IT_Pulse_Roue_A):
    """
    Callback incrémente le compteur d'impulsions, 12 par tour de roue (+ 12 par le capteur B) 
    
    """
    global compteur_roue
    compteur_roue = compteur_roue+1
   
#

def compteur2(IT_Pulse_Roue_B):
    """
    Callback incrémente le compteur d'impulsions, 12 par tour de roue (+12 par le capteur A)
    
    """
    global compteur_roue
    compteur_roue = compteur_roue+1    

#

##-----Les fonctions----##
##                      ##

def CmdPwm(pwm,rc) :
    """
    commande PWM LED BLEU
    commande en PWM des moteurs R et L
    pwm = pwmR ou pwmL ou pwmLED_BLEU 
    rc rapport cyclique de 0 à 1000 pour 0 à 100%
    
    """
    
    pwm.duty_u16(rc*2**16//1000)
    
#
    
def tourne() :
    """
    Le robot tourne sur lui même
    hor = 'h' sans horaire, 'a' anti-horaire
    Un tour complet = 51 impulsions
    """
    global hor
    if hor == 'h' :# horaire
        DirR .value(1)
        DirL .value(0)
    elif hor=='a' :# anti-horaire
        DirR .value(0)
        DirL .value(1)     
        
    rcr =150 # vitesse réduite en rotation
    CmdPwm (pwmR,rcr) # moteur droit
    rcl =150 # vitesse réduite en rotation       
    CmdPwm (pwmL,rcl) # moteur g

#
    
def stop () :
    """ Arrêt des moteurs
    """
    global compteur_roue
    global select
    global stop_correct
    
    # initialisations pret pour recevoir une nouvelle commande :
    rc = 0
    CmdPwm (pwmR,rc)
    DirR .value(0)
    rc = 0 
    CmdPwm (pwmL,rc)
    DirL .value(0)
    compteur_roue = 0
    select = 0
    stop_correct = 0
    CmdPwm(pwmLED_BLEU,5) # ralentissement cadence LED bleu
    
#

def ligne() :
    
    """
    Suivre une ligne noire en restant au centre de celle-ci avec correction automatique
     
    sl donne la position moyennée de la ligne noire par rapport au capteur à 8 photodiodes et 8 LEDS IR
    sl--> 0 à 8, 4.5 si piste noire centrée
    
    
    """
        
    
    #--> Détermination de sl :
    sn =0
    # affectation poids led(0) la plus à droite avec le poids le plus faible 1
    sl = 1*barled[0].value() + 2*barled[1].value() + 3*barled[2].value()+ 4*barled[3].value()+ 5*barled[4].value()+ 6*barled[5].value()+ 7*barled[6].value()+ 8*barled[7].value()
    # nombre de LED sur ligne noire
    sn = barled[0].value() + barled[1].value() + barled[2].value()+ barled[3].value()+ barled[4].value()+ barled[5].value()+ barled[6].value()+ barled[7].value()
    if sn  != 0 :
        sl = round(sl /sn ,1)# varie entre 0 et 8 --> 4.5 si piste noire centrée
    else:
        sl = consigne # pas de correction sur fond blanc, la ligne est considérée centrée
    #
    
    # en approche de la fin de ligne : pas de correction de trajectoire    
    if stop_correct == 1:
        sl = consigne # la ligne est centrée donc pas de correction 
        
     
    # 2 moteursconfiguration du sens de rotation avance directe :     
    DirR .value(0)
    DirL .value(0)
    
    #################################################################
    # A COMPLETER
    # variables disponibles:
    # s1
    # consigne
    # KP gain du proportionnel
    # vit0 vitesse du robot
    #
    # Objectif: calculer rcr et rcl les rapports cycliques des PWM= vitesse des roues droites et gauches
    # rcr et rcl doivent être des entiers, utiliser round()
    #################################################################
    #
    CmdPwm (pwmR,rcr) # commande en tension moteur droit
    CmdPwm (pwmL,rcl) # commande en tension moteur gauche


def select_fonction() :
    
    """ Selectionne la fonction d'ordre robot à éxecuter
        0 on passe, pas de mouvement du robot
        1 ligne, il parcourt une ligne
        2 rotation, il tourne
    """
   
    global select
     
    if select ==1:
        ligne()
    elif  select == 2:
        tourne()
        
 #
 
def filtre_message() :
    """ Traitement des commnades reçues par Bluetooth
        
    """
    
    global Fcommand
    global validCommand
    global select
    global cpt
    global hor
    global vit0
    global KP
    global stop_correct
    global compteur_roue
    # Traitement   des messages reçus par Bluetooth :
    
    if Fcommand[0] == 'b' : # mesure batterie
        send =str(round(pin_bat.read_u16() * facteur_bat,1))+'V'
        uart0.write('BAT :'+ send +'\r\n') 
        print('mesure bat', send)
        time.sleep(0.1)
         
         
    elif Fcommand[0] == 'd' : # envoi lecture barrette LED ligne
        send =''
        for i in range(7,-1,-1) : # mise dans l'ordre
            send =send +str(barled[i].value())
        uart0.write('BAR :'+ send +'\r\n')
        print('bar_led', send)
        time.sleep(0.1)
         
        
    elif Fcommand[0] == 'w' : # wait ou stop -> initialisation
        stop()
        select =0
        
    #    
    elif Fcommand[0] == 'l' : # parcours ligne
        select =1
        CmdPwm(pwmLED_BLEU,100) # signalisation par rythme battement LED bleu
        stop_correct =0
        compteur_roue = 0
        try :
            cpt = int(Fcommand[1:5]) # 4 chiffres compteur_roue
        except ValueError:
            cpt = 0
    elif Fcommand[0] == 'v' : # programmer vitesse de base  robot
               
        try :  
            vit0= int(Fcommand[1:4]) # 4 chiffres compteur_roue
        except ValueError:
            vit0 = vitdefaut

    elif Fcommand[0] == 'k' : # programmer KP robot
        #select =0
        
        try :
            KP= int(Fcommand[1:4]) # 4 chiffres compteur_roue
        except ValueError:
            KP = KPdefaut   

     
    elif Fcommand[0] == 'c' : # cercle
        select =2
        CmdPwm(pwmLED_BLEU,100) # signalisation par rythme battement LED bleu
        stop_correct =0
        compteur_roue = 0
        if Fcommand[1]  =='h': # sens horaire
            hor ='h'
        elif Fcommand[1]  =='a': # sens anti-horaire
            hor ='a'  
        try :
            cpt = int(Fcommand[2:6]) # 4 chiffres compteur_roue
        except ValueError:
            cpt = 0
   
 
##-----Début programme----##
##                        ##
            
IT_Pulse_Roue_A.irq(trigger=Pin.IRQ_RISING, handler=compteur) # active irq entrée pulse capteur roue gauche
IT_Pulse_Roue_B.irq(trigger=Pin.IRQ_FALLING, handler=compteur2) # active irq entrée pulse capteur roue gauche

stop() # Arrêt moteurs et initialisations
 
time.sleep(1)
 
uart0.write('Start\r\n') # premier message vers Bluetooth 'Start'
time.sleep(1)
print('ROBOT OK')

##
while True :
    time.sleep(0.05) 
    #Time Out Alimentation mise OFF
    if cpt_kill == cpt_kill_max :
        kill_Pin .value(1) # arrêt alim.
        print('!!!Stop Time OUT Alimentation!!!')
        print('Rallumer le robot avec son bouton rouge')
        # ces deux lignes permettent de redémarrer si on est connecté à Thonny
        time.sleep(1)
        kill_Pin .value(0)
     #   
    select_fonction() # traitement mouvement en ligne ou rotation ou rien
    
    # Fin de ligne ou rotation atteinte? :
    #
    if compteur_roue >= cpt : # valeur compteur_roue atteinte
        if (select == 1) or (select == 2) :# mode ligne ou rotation
            
            stop()
            uart0.write('complete'+'\r\n') # opération traité même rotation cx0000 et l0000
            print('complete')
            #éteindre LED bleu d'état
    #        
    # Décision d'arrêt de correction de suivi de ligne 1/2 tour de roue avant l'arrivée
    # en ayant effectué au moins 1/2 tours de roue
    #       
    if (select ==1) and (compteur_roue >0.5 *tourRoue) and  (compteur_roue > cpt-0.5* tourRoue):
        stop_correct =1
        
    #  
    # Réception des commandes Bluetooth via le port COM0 :
    # traitement de la validité au niveau du format
    if uart0.any():
        
        # mise à 0 du compteur Time Out d'alimentation
        cpt_kill = 0
        
        time.sleep(0.1)
        command = uart0.readline()
        Fcommand=  ((str(command,'utf-8'))[:-2]).lstrip() # réception en octet (binaire) à convertir en string,suppression des caractéres vides transmis pour synchro
        k = len(Fcommand) # longueur de message valide pour le slicing
#        if (k == 1)or ...  (k == 6) : # 1 pour w et b,d, 6 pour caxxxx et chxxxx, 5 pour lxxxx, 4 pour vxxx
        
        if (k == 1) and Fcommand in LC : # commande d'un caractère reçu contenu dans la liste LC    
 
            validCommand = 1
            
            
        elif (k == 4) and  (Fcommand[0] == 'v' or Fcommand[0] == 'k') :
            if ('/' < Fcommand[1] < ':') and ('/' < Fcommand[2] < ':') and ('/' < Fcommand[3] < ':' ) :
 
                validCommand = 1
 
            
        elif (k == 5) and  Fcommand[0] == 'l' :
            if ('/' < Fcommand[1] < ':') and ('/' < Fcommand[2] < ':') and ('/' < Fcommand[3] < ':') and ('/' < Fcommand[4] < ':') :
 
                validCommand = 1
 
                    
        elif (k == 6) and  (Fcommand[0:2] == 'ch' or Fcommand[0:2] == 'ca') :
            if ('/' < Fcommand[2] < ':') and ('/' < Fcommand[3] < ':') and ('/' < Fcommand[4] < ':') and ('/' < Fcommand[5] < ':') :
 
                validCommand = 1
        else :
                      
            validCommand = 0
            
        # Retour commande reçue valide ou pas ?
        # Acquitement par envoi ok ou Ko
        # demande affichage barrette LED ou mesure batterie pas de ok envoyé
        if validCommand == 1 :
            print('commande valide:   ',Fcommand)
            filtre_message()        
            validCommand = 0
            if not( Fcommand[0] == 'b'or Fcommand[0] == 'd'): # si demande affichage barrette LED ou mesure batterie pas de ok
                
                uart0.write('ok'+'\r\n')
                time.sleep(0.1)
            
        else :
            command ='  '
            validCommand = 0
            uart0.write('ko'+'\r\n')
            time.sleep(0.1)
            print('erreur commande:',Fcommand)
        
# Fin programme   


