########################################################################################################################
#
# Module: traitement Dijkstra
#
# fonctions :
# dijkstra(graphe, source='A') : Algorithme de Dijkstra
#
# calc_Dijkstra(source='A', destination='P') : calcul le parcours du robot, de la source à l'arrivée
#
# exec_Dijk() : lance l'éxécution du parcours par le robot
########################################################################################################################


from data_graphe import graphe
from carte_parcours_robot import calc_trajet
from envoi_robot import execution

def noeudpluspetitlambda(liste_a_traiter):
    #liste_a_traiter contient la liste des noeuds à traiter
    #chaque element de liste contient: x (noeud), lambda_i[x] (meilleure longueur connue pour aller à x), q[x] (meilleur prédécesseur), m[x] (marque)
    #a_traiter = [(lambda_1, noeud1, q_1, m_1), (lambda_2, noeud 2, q_2, m_2), ...]
    #pour afficher la liste : 
    #print(liste_a_traiter)
    #acceder au premier element de la liste liste_a_traiter[0]
    
    ####################################################################################
    #                     A MODIFIER
    # initialisation 
    plus_petit_lambda = liste_a_traiter[0][1]
    noeud = liste_a_traiter[0][0]
    ####################################################################################
    #  completer ici pour parcourir la liste et trouver noeud avec le plus petit lambda, 
    #  non marqué
    #
    ####################################################################################
    
    
    return plus_petit_lambda, noeud 

def dijkstra(graphe, source='A'):
    """ Cette fonction traite l'algorithme de Dijkstra"""

    # test validité, couts positifs (arcs reliant les sommets U ={ x1,x1...xn})
    assert all(graphe[u][v] >= 0 for u in graphe.keys() for v in graphe[u].keys())
    # Initialisation des 3 dictionnaires : q les précédents à None, les sommets à False et lambda_i à inf #
    q = {x: None for x in graphe.keys()}  # sommets précédents None  q={'A':None,'B':None ... }
    m = {x: False for x in graphe.keys()}  # marquage  des sommets False m={'A':False,'B':False ... }
    lambda_i = {x: float('inf') for x in graphe.keys()}  # lambda_i = {A: float('inf'),'B':float('inf') ... }
    lambda_i[source] = 0  # exemple si source est A, lambda_i = {A: 0,'B':inf .....}
    
    #creation de la liste a_traiter avec tous les noeuds, leur lambda_i, q_i et m_i
    a_traiter = []
    for x in graphe.keys():
        a_traiter.append((x, lambda_i[x], q[x], m[x]))
    print(a_traiter)    
    

    while a_traiter:  # jusqu'à a_traiter soit  False donc vide []
        # on cherche dans a_traiter le noeud avec le plus petit lambda
        lambda_sommet, sommet = noeudpluspetitlambda(a_traiter)
        # on retire de a_traiter le noeud avec le plus petit lambda
        a_traiter.remove((sommet, lambda_i[sommet], q[sommet], m[sommet]))
        
        if not m[sommet]: #inutile a priori
            m[sommet] = True  # sommet est marqué
            for voisin in graphe[sommet].keys():  # On pioche dans les voisins x_j du sommet marqué précédent
                ####################################################################################
                #                     A MODIFIER ET CHANGER
                # sommet : le sommet à traiter
                # voisin : sommet voisin
                # round(x, 2) : arrondi x à deux chiffres après la virgule (pour le calcul de coût)
                # graphe[x_i][x_j] : donne le coût entre x_i et x_j dans le graphe
                #
                # attention : si mise à jour, veiller à effacer de la liste a traiter le noeud "voisin"
                # et à le remettre dans la liste.
                # fonctions a_traiter.remove et a_traiter.append
                ####################################################################################
                    
                # print(a_traiter)
        
    return lambda_i, q  # lambda_i et meilleur prédécesseur


def calc_Dijkstra(source='A', destination='P'):
    """Cette fonction retourne les informations  d'un parcours exécutable par le robot,
     de la source à l'arrivéé, 3 listes :
     vit_segment, depl_segment et rot_robot pour les vitesses, mouvements en ligne droite et en rotation"""

    global vit_segment, depl_segment, rot_robot
    dp = destination

    print("Les voisins d'un point Px et leur cout par rapport à Px:")
    for aff in graphe.keys():
        print(aff, ' : ', graphe[aff])

    lambda_i, q = dijkstra(graphe, source)  # dijkstra(graphe,'source') La source par défaut est 'A'
    print('En partant de la source --> : ', source)
    print('Cout minimum  pour aller à:', lambda_i)
    print('Liste des précédents :', q)
    print("------>")
    parcours = [destination]  # dernier élémént du parcours
    print('Cout minimum pour aller à', dp, ' :', lambda_i[dp])
    print('en passant par les points ', end='')
    # reconstruction du chemin à l'endroit

    while q[dp] != None:
        dp = q[dp]
        parcours.append(dp)  # reconstruction du chemin à l'endroit
    parcours.reverse()

    if q[destination] == None:  # cas ou pas de chemin possible
        parcours = []

    print('du parcours', parcours)
    textinfo1 = 'Le cout minimum pour aller à ' + destination + ' : ' + str(lambda_i[destination])
    textinfo2 = 'en passant par les  points : ' + '-'.join(str(x) for x in parcours)

    # calcul des vitesse des lignes et rotationd de déplacement du robot si parcours exiqte
    if not (parcours == []):
        vit_segment, depl_segment, rot_robot = calc_trajet(parcours)
        print(vit_segment, depl_segment, rot_robot)

    return textinfo1, textinfo2  # retour info vers l'interface graphique


def exec_Dijk():
    """Cette fonction lance l'éxécution du parcours par le robot"""

    global vit_segment, depl_segment, rot_robot

    execution(vit_segment, depl_segment, rot_robot)
